/*
        printf("New job dispatched\n");
 * Copyright 2010 Jeff Garzik
 * Copyright 2012-2015 pooler
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.  See COPYING for more details.
 */

#include "sfminer-config.h"
#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <inttypes.h>
#include <unistd.h>
#include <sys/time.h>
#include <time.h>
#ifdef WIN32
#include <windows.h>
#else
#include <errno.h>
#include <signal.h>
#include <sys/resource.h>
#if HAVE_SYS_SYSCTL_H
#include <sys/types.h>
#if HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif
#include <sys/sysctl.h>
#endif
#endif
#include <jansson.h>
#include <curl/curl.h>
#include "compat.h"
#include "miner.h"
#include "sfards.h"

#define PROGRAM_NAME        "minerd"

#define MAX_NTIME_OFFSET         (32)
#define MAX_JOB_ID_LENGTH        (56)
#define STATISTC_INTERVAL        (300)

#ifdef __linux /* Linux specific policy and affinity management */
#include <sched.h>
static inline void drop_policy(void)
{
    struct sched_param param;
    param.sched_priority = 0;

#ifdef SCHED_IDLE
    if (unlikely(sched_setscheduler(0, SCHED_IDLE, &param) == -1))
#endif
#ifdef SCHED_BATCH
        sched_setscheduler(0, SCHED_BATCH, &param);
#endif
}

static inline void affine_to_cpu(int id, int cpu)
{
    cpu_set_t set;

    CPU_ZERO(&set);
    CPU_SET(cpu, &set);
    sched_setaffinity(0, sizeof(set), &set);
}
#elif defined(__FreeBSD__) /* FreeBSD specific policy and affinity management */
#include <sys/cpuset.h>
static inline void drop_policy(void)
{
}

static inline void affine_to_cpu(int id, int cpu)
{
    cpuset_t set;
    CPU_ZERO(&set);
    CPU_SET(cpu, &set);
    cpuset_setaffinity(CPU_LEVEL_WHICH, CPU_WHICH_TID, -1, sizeof(cpuset_t), &set);
}
#else
static inline void drop_policy(void)
{
}

static inline void affine_to_cpu(int id, int cpu)
{
}
#endif
        
enum workio_commands 
{
    WC_GET_WORK,
    WC_SUBMIT_WORK,
};

struct workio_cmd 
{
    enum workio_commands    cmd;
    struct thr_info        *thr;
    union 
    {
        struct work    *work;
    } u;
};

enum algos 
{
    ALGO_SCRYPT,        /* scrypt(1024,1,1) */
    ALGO_SHA256D,        /* SHA-256d */
};

static const char *algo_names[] = 
{
    [ALGO_SCRYPT]  = "scrypt",
    [ALGO_SHA256D] = "sha256d",
};

struct sfards_dev *sfards_dev = NULL;
static int sfards_freq = 25;
static char sfards_paths[256] = {0,};

bool opt_debug = false;
bool opt_protocol = false;
static bool opt_benchmark = false;
bool opt_redirect = true;
bool want_longpoll = true;
bool have_longpoll = false;
bool have_gbt = true;
bool allow_getwork = true;
bool want_stratum = true;
bool have_stratum = false;
bool use_syslog = false;
static bool opt_background = false;
static bool opt_quiet = false;
static int opt_retries = -1;
static int opt_fail_pause = 30;
int opt_timeout = 0;
static int opt_scantime = 5;
static const bool opt_time = true;
static enum algos opt_algo = ALGO_SCRYPT;
static int opt_scrypt_n = 1024;
static int opt_n_threads;
static int num_processors;
static char *rpc_url;
static char *rpc_userpass;
static char *rpc_user, *rpc_pass;
static int pk_script_size;
static unsigned char pk_script[25];
static char coinbase_sig[101] = "";
char *opt_cert;
char *opt_proxy;
long opt_proxy_type;
struct thr_info *thr_info;
static int work_thr_id;
static int workfifo_thr_id;
int longpoll_thr_id = -1;
int stratum_thr_id = -1;
struct work_restart *work_restart = NULL;
static struct stratum_ctx stratum;
static struct workfifo workfifo;

pthread_mutex_t applog_lock;
static pthread_mutex_t stats_lock;

static unsigned long accepted_count = 0L;
static unsigned long rejected_count = 0L;
static unsigned long unsubmited_count = 0L;
static unsigned long submited_count = 0L;
static time_t start_time = 0;
static double *thr_hashrates;

#ifdef HAVE_GETOPT_LONG
#include <getopt.h>
#else
struct option {
    const char *name;
    int has_arg;
    int *flag;
    int val;
};
#endif

static char const usage[] = "\
Usage: " PROGRAM_NAME " [OPTIONS]\n\
Options:\n\
  -a, --algo=ALGO       specify the algorithm to use\n\
                          scrypt    scrypt(1024, 1, 1) (default)\n\
                          scrypt:N  scrypt(N, 1, 1)\n\
                          sha256d   SHA-256d\n\
  -o, --url=URL         URL of mining server\n\
  -O, --userpass=U:P    username:password pair for mining server\n\
  -u, --user=USERNAME   username for mining server\n\
  -p, --pass=PASSWORD   password for mining server\n\
      --cert=FILE       certificate for mining server using SSL\n\
  -x, --proxy=[PROTOCOL://]HOST[:PORT]  connect through a proxy\n\
  -t, --threads=N       number of miner threads (default: number of processors)\n\
  -r, --retries=N       number of times to retry if a network call fails\n\
                          (default: retry indefinitely)\n\
  -R, --retry-pause=N   time to pause between retries, in seconds (default: 30)\n\
  -T, --timeout=N       timeout for long polling, in seconds (default: none)\n\
  -s, --scantime=N      upper bound on time spent scanning current work when\n\
                          long polling is unavailable, in seconds (default: 5)\n\
      --coinbase-addr=ADDR  payout address for solo mining\n\
      --coinbase-sig=TEXT  data to insert in the coinbase when possible\n\
      --no-longpoll     disable long polling support\n\
      --no-getwork      disable getwork support\n\
      --no-gbt          disable getblocktemplate support\n\
      --no-stratum      disable X-Stratum support\n\
      --no-redirect     ignore requests to change the URL of the mining server\n\
  -q, --quiet           disable per-thread hashmeter output\n\
  -D, --debug           enable debug output\n\
  -P, --protocol-dump   verbose dump of protocol-level activities\n\
  -d, --device        path of serial port(sfards device only)\n\
  -f, --freq        frequency of SF3301(sfards device only)\n"  
#ifdef HAVE_SYSLOG_H
"\
  -S, --syslog          use system log for output messages\n"
#endif
#ifndef WIN32
"\
  -B, --background      run the miner in the background\n"
#endif
"\
      --benchmark       run in offline benchmark mode\n\
  -c, --config=FILE     load a JSON-format configuration file\n\
  -V, --version         display version information and exit\n\
  -h, --help            display this help text and exit\n\
";

static char const short_options[] =
#ifndef WIN32
    "B"
#endif
#ifdef HAVE_SYSLOG_H
    "S"
#endif
    "a:c:Dhp:Px:qr:R:s:t:T:o:u:O:Vf:d:";

static struct option const options[] = {
    { "algo", 1, NULL, 'a' },
        { "device", 1, NULL, 'd' },
        { "freq", 1, NULL, 'f' },
#ifndef WIN32
    { "background", 0, NULL, 'B' },
#endif
    { "benchmark", 0, NULL, 1005 },
    { "cert", 1, NULL, 1001 },
    { "coinbase-addr", 1, NULL, 1013 },
    { "coinbase-sig", 1, NULL, 1015 },
    { "config", 1, NULL, 'c' },
    { "debug", 0, NULL, 'D' },
    { "help", 0, NULL, 'h' },
    { "no-gbt", 0, NULL, 1011 },
    { "no-getwork", 0, NULL, 1010 },
    { "no-longpoll", 0, NULL, 1003 },
    { "no-redirect", 0, NULL, 1009 },
    { "no-stratum", 0, NULL, 1007 },
    { "pass", 1, NULL, 'p' },
    { "protocol-dump", 0, NULL, 'P' },
    { "proxy", 1, NULL, 'x' },
    { "quiet", 0, NULL, 'q' },
    { "retries", 1, NULL, 'r' },
    { "retry-pause", 1, NULL, 'R' },
    { "scantime", 1, NULL, 's' },
#ifdef HAVE_SYSLOG_H
    { "syslog", 0, NULL, 'S' },
#endif
    { "threads", 1, NULL, 't' },
    { "timeout", 1, NULL, 'T' },
    { "url", 1, NULL, 'o' },
    { "user", 1, NULL, 'u' },
    { "userpass", 1, NULL, 'O' },
    { "version", 0, NULL, 'V' },
    { 0, 0, 0, 0 }
};

struct work 
{
    uint32_t data[32];
    uint32_t target[8];
    uint32_t midstate[8];
    size_t xnonce2_len;
    unsigned char xnonce2[4];
    char job_id[MAX_JOB_ID_LENGTH];
};

#define WORKFIFO_NOTIFY_FLUSH            (0)
#define WORKFIFO_NOTIFY_LOAD            (1)
#define WORKFIFO_NOTIFY_EXIT            (2)
#define WORKFIFO_SIZE                (16)
#define WORKFIFO_MASK                (WORKFIFO_SIZE - 1)
#define WORKFIFO_COUNT(wf)            (((wf)->widx + WORKFIFO_SIZE - (wf)->ridx) & WORKFIFO_MASK)
#define WORKFIFO_FULL(wf)            ((((wf)->widx + 1) & WORKFIFO_MASK) == (wf)->ridx)
#define WORKFIFO_EMPTY(wf)            ((wf)->widx  == (wf)->ridx)
#define WORKFIFO_READ(wf)            ((wf)->ridx = ((wf)->ridx + 1) & WORKFIFO_MASK)
#define WORKFIFO_WRITE(wf)            ((wf)->widx = ((wf)->widx + 1) & WORKFIFO_MASK)

struct workfifo 
{
    struct work work[WORKFIFO_SIZE];
    int ridx;
    int widx;
    char job_id[MAX_JOB_ID_LENGTH];
    int notify;
    pthread_mutex_t lock;
    pthread_cond_t cond;    
};


static inline void work_copy(struct work *dest, const struct work *src)
{
    memcpy(dest, src, sizeof(struct work));
}

static void workfifo_init(struct workfifo *wf)
{
    memset(wf,0,sizeof(*wf));
    pthread_mutex_init(&wf->lock,NULL);
    pthread_cond_init(&wf->cond,NULL);
}

static void workfifo_destroy(struct workfifo *wf)
{
    pthread_mutex_destroy(&wf->lock);
    pthread_cond_destroy(&wf->cond);
}

static void workfifo_notify(struct workfifo *wf,int notify)
{
    pthread_mutex_lock(&wf->lock);
    
    wf->notify |= (1 << notify);

    pthread_mutex_unlock(&wf->lock);

    pthread_cond_signal(&wf->cond);
}

static int workfifo_fetch(struct workfifo *wf,struct work *work)
{
    int count;
    pthread_mutex_lock(&wf->lock);
    
    if (WORKFIFO_EMPTY(wf)) 
    {
        pthread_mutex_unlock(&wf->lock);
        workfifo_notify(wf,WORKFIFO_NOTIFY_LOAD);
        return 1;
    }
    
    work_copy(work,&wf->work[wf->ridx]);
    WORKFIFO_READ(wf);
    count = WORKFIFO_COUNT(wf);

    pthread_mutex_unlock(&wf->lock);
    
    if (count < (WORKFIFO_SIZE >> 1)) 
    {
        workfifo_notify(wf,WORKFIFO_NOTIFY_LOAD);
    }

    return 0;
}

static void stratum_gen_work(struct stratum_ctx *sctx, struct work *work);

static void *workfifo_thread(void *userdata)
{
    struct workfifo *wf = &workfifo;
    
    pthread_mutex_lock(&wf->lock);

    for (;;) 
    {
        if (wf->notify)    
        {
            if (wf->notify & (1 << WORKFIFO_NOTIFY_EXIT)) 
            {
                break;
            } 
            else if (wf->notify & (1 << WORKFIFO_NOTIFY_FLUSH)) 
            {
                wf->widx = wf->ridx;
            }
            wf->notify = 0;
        }
        else 
        {
            pthread_cond_wait(&wf->cond,&wf->lock);
            continue;
        } 
        
        pthread_mutex_unlock(&wf->lock);

        while (strlen(stratum.job.job_id) && !WORKFIFO_FULL(wf) && wf->notify == 0) 
        {
            stratum_gen_work(&stratum, &wf->work[wf->widx]);
            WORKFIFO_WRITE(wf);            
        }        
        
        pthread_mutex_lock(&wf->lock);
    }
    
    pthread_mutex_unlock(&wf->lock);
    return NULL;
}

static void share_result(int result, const char *reason)
{
    char s[345];
    double hashrate;
    int i;

    hashrate = 0.;
    pthread_mutex_lock(&stats_lock);
    for (i = 0; i < opt_n_threads; i++)
        hashrate += thr_hashrates[i];
    result ? accepted_count++ : rejected_count++;
    pthread_mutex_unlock(&stats_lock);
    
    sprintf(s, hashrate >= 1e6 ? "%.0f" : "%.2f", 1e-3 * hashrate);
/*
    applog(LOG_INFO, "accepted: %lu/%lu (%.2f%%), %s khash/s %s",
           accepted_count,
           accepted_count + rejected_count,
           100. * accepted_count / (accepted_count + rejected_count),
           s,
           result ? "(yay!!!)" : "(booooo)");
*/
    applog(LOG_INFO, "accepted: %lu/%lu (%.2f%%), %s   --> submited: %lu/%lu/%lu",
           accepted_count,
           accepted_count + rejected_count,
           100. * accepted_count / (accepted_count + rejected_count),
           result ? "(yay!!!)" : "(booooo)",
           accepted_count + rejected_count,unsubmited_count,submited_count);
    
    if (reason)
            applog(LOG_INFO, "DEBUG: reject reason: %s", reason);
    if (opt_debug && reason)
        applog(LOG_DEBUG, "DEBUG: reject reason: %s", reason);
}

static bool submit_upstream_work(CURL *curl, struct work *work)
{
    json_t *val, *res, *reason;
    char data_str[2 * sizeof(work->data) + 1];
    char s[345];
    int i;
    bool rc = false;
    /* pass if the previous hash is not the current previous hash */
    uint32_t ntime, nonce;
    char ntimestr[9], noncestr[9], *xnonce2str, *req;

    le32enc(&ntime, work->data[17]);
    le32enc(&nonce, work->data[19]);
    bin2hex(ntimestr, (const unsigned char *)(&ntime), 4);
    bin2hex(noncestr, (const unsigned char *)(&nonce), 4);
    xnonce2str = malloc(work->xnonce2_len * 2 + 1);
    bin2hex(xnonce2str,work->xnonce2,work->xnonce2_len < 4 ? work->xnonce2_len : 4);
    for (i = 8;i < work->xnonce2_len * 2;i++)
    {
        xnonce2str[i] = '0';        
    }
    xnonce2str[work->xnonce2_len * 2] = '\0';
 
    req = malloc(256 + strlen(rpc_user) + strlen(work->job_id) + 2 * work->xnonce2_len);
    sprintf(req,"{\"method\": \"mining.submit\", \"params\": [\"%s\", \"%s\", \"%s\", \"%s\", \"%s\"], \"id\":4}",
                rpc_user, work->job_id, xnonce2str, ntimestr, noncestr);

    if (opt_debug)
    {
        applog(LOG_DEBUG, "DEBUG:submit job_id='%s' extranonce2=%s ntime=%08x nonce=%08x",
               work->job_id, xnonce2str, swab32(work->data[17]),work->data[19]);
    }

    free(xnonce2str);
    rc = stratum_send_line(&stratum, req);
    free(req);
    if (unlikely(!rc)) 
    {
        applog(LOG_ERR, "submit_upstream_work stratum_send_line failed");
        goto out;
    }
    rc = true;

out:
    return rc;
}

static const char *getwork_req =
    "{\"method\": \"getwork\", \"params\": [], \"id\":0}\r\n";

#define GBT_CAPABILITIES "[\"coinbasetxn\", \"coinbasevalue\", \"longpoll\", \"workid\"]"

static const char *gbt_req =
    "{\"method\": \"getblocktemplate\", \"params\": [{\"capabilities\": "
    GBT_CAPABILITIES "}], \"id\":0}\r\n";
static const char *gbt_lp_req =
    "{\"method\": \"getblocktemplate\", \"params\": [{\"capabilities\": "
    GBT_CAPABILITIES ", \"longpollid\": \"%s\"}], \"id\":0}\r\n";

static bool get_upstream_work(CURL *curl, struct work *work)
{
    return true;
}

static void workio_cmd_free(struct workio_cmd *wc)
{
    if (!wc)
    {
        return;
    }
    switch (wc->cmd) 
    {
    case WC_SUBMIT_WORK:
        free(wc->u.work);
        break;    
    default: /* do nothing */
    break;
    }

    memset(wc, 0, sizeof(*wc));    /* poison */
    free(wc);
}

static bool workio_get_work(struct workio_cmd *wc, CURL *curl)
{
    struct work *ret_work;
    int failures = 0;

    ret_work = calloc(1, sizeof(*ret_work));
    if (!ret_work)
        return false;

    /* obtain new work from bitcoin via JSON-RPC */
    while (!get_upstream_work(curl, ret_work)) 
    {
        if (unlikely((opt_retries >= 0) && (++failures > opt_retries))) 
        {
            applog(LOG_ERR, "json_rpc_call failed, terminating workio thread");
            free(ret_work);
            return false;
        }

        /* pause, then restart work-request loop */
        applog(LOG_ERR, "json_rpc_call failed, retry after %d seconds",
            opt_fail_pause);
        sleep(opt_fail_pause);
    }

    /* send work to requesting thread */
    if (!tq_push(wc->thr->q, ret_work))
        free(ret_work);

    return true;
}

static bool workio_submit_work(struct workio_cmd *wc, CURL *curl)
{
    int failures = 0;

    /* submit solution to bitcoin via JSON-RPC */
    
    while (!submit_upstream_work(curl, wc->u.work)) 
    {
        if (unlikely((opt_retries >= 0) && (++failures > opt_retries))) 
        {
            applog(LOG_ERR, "...terminating workio thread");
            return false;
        }

        /* pause, then restart work-request loop */
        applog(LOG_ERR, "...retry after %d seconds",opt_fail_pause);
        sleep(opt_fail_pause);
    }

    return true;
}

static void *workio_thread(void *userdata)
{
    struct thr_info *mythr = userdata;
    CURL *curl;
    bool ok = true;

    curl = curl_easy_init();
    if (unlikely(!curl)) 
    {
        applog(LOG_ERR, "CURL initialization failed");
        return NULL;
    }

    while (ok) 
    {
        struct workio_cmd *wc;

        /* wait for workio_cmd sent to us, on our queue */
        wc = tq_pop(mythr->q, NULL);
        if (!wc) 
        {
            ok = false;
            break;
        }

        /* process workio_cmd */
        switch (wc->cmd) {
        case WC_GET_WORK:
            ok = workio_get_work(wc, curl);
            break;
        case WC_SUBMIT_WORK:
            ok = workio_submit_work(wc, curl);
            break;
        default:        /* should never happen */
            ok = false;
            break;
        }

        workio_cmd_free(wc);
    }

    tq_freeze(mythr->q);
    curl_easy_cleanup(curl);

    return NULL;
}

static bool get_work(struct thr_info *thr, struct work *work)
{
    struct workio_cmd *wc;
    struct work *work_heap;

    if (opt_benchmark) {
        memset(work->data, 0x55, 76);
        work->data[17] = swab32(time(NULL));
        memset(work->data + 19, 0x00, 52);
        work->data[20] = 0x80000000;
        work->data[31] = 0x00000280;
        memset(work->target, 0x00, sizeof(work->target));
        return true;
    }

    /* fill out work request message */
    wc = calloc(1, sizeof(*wc));
    if (!wc)
        return false;

    wc->cmd = WC_GET_WORK;
    wc->thr = thr;

    /* send work request to workio thread */
    if (!tq_push(thr_info[work_thr_id].q, wc)) 
    {
        workio_cmd_free(wc);
        return false;
    }

    /* wait for response, a unit of work */
    work_heap = tq_pop(thr->q, NULL);
    if (!work_heap)
        return false;

    /* copy returned work into storage provided by caller */
    memcpy(work, work_heap, sizeof(*work));
    free(work_heap);

    return true;
}

static bool submit_work(struct thr_info *thr,struct work *work_in)
{
    struct workio_cmd *wc;
    
    /* fill out work request message */
    wc = calloc(1, sizeof(*wc));
    if (!wc)
    {
        free(work_in);
        return false;
    }
    wc->cmd = WC_SUBMIT_WORK;
    wc->thr = thr;
    wc->u.work = work_in;
    
    /* send solution to workio thread */
    if (!tq_push(thr_info[work_thr_id].q, wc))
    {
        goto err_out;
    }

    return true;

err_out:
    workio_cmd_free(wc);
    return false;
}

static void stratum_gen_work(struct stratum_ctx *sctx, struct work *work)
{
    unsigned char merkle_root[64];
    int i;

    pthread_mutex_lock(&sctx->work_lock);
    
    strcpy(work->job_id,sctx->job.job_id);
    work->xnonce2_len = sctx->xnonce2_size;
    memcpy(work->xnonce2,sctx->job.xnonce2,sctx->job.xnonce2_len); 

    /* Generate merkle root */
    sha256d(merkle_root, sctx->job.coinbase, sctx->job.coinbase_size);
    for (i = 0; i < sctx->job.merkle_count; i++) 
    {
        memcpy(merkle_root + 32, sctx->job.merkle[i], 32);
        sha256d(merkle_root, merkle_root, 64);
    }
    
    /* Increment extranonce2 */
    for (i = 0; i < sctx->job.xnonce2_len && !++sctx->job.xnonce2[i]; i++);

    /* Assemble block header */
    memset(work->data, 0, 128);
    work->data[0] = le32dec(sctx->job.version);
    for (i = 0; i < 8; i++)
    {
        work->data[1 + i] = le32dec((uint32_t *)sctx->job.prevhash + i);
    }

    for (i = 0; i < 8; i++)
    {
        work->data[9 + i] = be32dec((uint32_t *)merkle_root + i);
    }
    work->data[17] = le32dec(sctx->job.ntime);
    work->data[18] = le32dec(sctx->job.nbits);
    work->data[20] = 0x80000000;
    work->data[31] = 0x00000280;

    sha256d_midstate(work->midstate,work->data);
    pthread_mutex_unlock(&sctx->work_lock);

    if (opt_debug) 
    {
        char *xnonce2str = abin2hex(work->xnonce2, sctx->job.xnonce2_len);
        applog(LOG_DEBUG, "DEBUG:generate job_id='%s' extranonce2=%s ntime=%08x",
               work->job_id, xnonce2str, swab32(work->data[17]));
        free(xnonce2str);
    }

    if (opt_algo == ALGO_SCRYPT)
    {
        diff_to_target(work->target, sctx->job.diff / 65536.0);
    }
    else
    {
        diff_to_target(work->target, sctx->job.diff);
    }
}

static int stratum_port_status(struct sfards_port *port)
{
    int nonce_n[PORT_MAX_CHIPID];
    int err_n[PORT_MAX_CHIPID];
    int freq[PORT_MAX_CHIPID];
    float hashrate[PORT_MAX_CHIPID];
    int sum_nonce_n,sum_err_n,sum_freq,i;
    float sum_hashrate,correct;
    char str_nonce[64] = {0,};
    char str_err[64] = {0,};
    char str_freq[64] = {0,};
    char str_hashrate[64] = {0,};

    sfards_port_status_get(port,PORT_MAX_CHIPID,nonce_n,err_n,freq,hashrate);

    sum_freq = 0;
    sum_nonce_n = 0;
    sum_err_n = 0;
    sum_hashrate = 0.0;

    for (i = 0;i < PORT_MAX_CHIPID;i++)
    {
        sum_nonce_n += nonce_n[i];
        sum_err_n += err_n[i];
        sum_hashrate += hashrate[i];
        sprintf(str_nonce,"%s %6d",str_nonce,nonce_n[i]);
        sprintf(str_err,"%s %6d",str_err,err_n[i]);
        sprintf(str_freq,"%s %6d",str_freq,freq[i]);
        if (port->type == PORT_TYPE_SCRYPT)
        {
            sprintf(str_hashrate,"%s %6.2f",str_hashrate,hashrate[i]);
        } 
        else
        {
            sprintf(str_hashrate,"%s %6.2f",str_hashrate,hashrate[i] / 1000.0);
        }
        sum_freq += freq[i];
    }

    correct = (sum_nonce_n + sum_err_n != 0) 
              ? sum_nonce_n * 100.0 / (sum_nonce_n + sum_err_n) : 0.0;
    if (port->type == PORT_TYPE_SCRYPT)
    {
        applog(LOG_INFO,"PORT %s : nonce = %d(%3.1f%%) err = %d hashrate = %6.2f Mhash",
                        port->path,sum_nonce_n,correct,sum_err_n,sum_hashrate); 
        port->hashrate = ((float)(sum_freq * 124)) / (23 * 2 * 1000);
    }
    else
    {
        applog(LOG_INFO,"PORT %s : nonce = %d(%3.1f%%) err = %d hashrate = %6.2f Ghash",
                        port->path,sum_nonce_n,correct,sum_err_n,sum_hashrate / 1000.0); 
        port->hashrate = (float)(sum_freq * 160);
    }
    applog(LOG_INFO,"    Nonce    : [%s]",str_nonce);
    applog(LOG_INFO,"    Error    : [%s]",str_err);
    applog(LOG_INFO,"    Freq     : [%s]",str_freq);
    applog(LOG_INFO,"    Hashrate : [%s]",str_hashrate);   

    port->errors = sum_err_n;
}

static inline void stratum_submit_work(struct thr_info *mythr,struct work *work,
                                       uint32_t offset,uint32_t nonce)
{
    struct work *submit = malloc(sizeof(struct work));
    if (!submit)
    {
        pthread_mutex_lock(&stats_lock);
        unsubmited_count++;
        pthread_mutex_unlock(&stats_lock);
    }
    else 
    {
        work_copy(submit,work);
        submit->data[19] = nonce;
        submit->data[17] = swab32(swab32(submit->data[17]) + offset); 
        if (!opt_benchmark && !submit_work(mythr, submit))
        {
            pthread_mutex_lock(&stats_lock);
            unsubmited_count++;
            pthread_mutex_unlock(&stats_lock);
        }
        else
        {
            pthread_mutex_lock(&stats_lock);
            submited_count++;
            pthread_mutex_unlock(&stats_lock);
        }
    }
}

static void *stratum_scrypt_thread(void *userdata)
{
    struct thr_info *mythr = userdata;
    struct sfards_port *port;
    struct work work = {{0}};
    struct sfards_rpt nonce_rpt[8];
    struct timespec ts;
    time_t stat_to;
    unsigned char *scratchbuf = NULL;
    int thr_id;
    int i,count;

    thr_id = mythr->id;
    port = &sfards_dev->port[thr_id];
    /* Set worker threads to nice 19 and then preferentially to SCHED_IDLE
     * and if that fails, then SCHED_BATCH. No need for this to be an
     * error if it fails */
    if (!opt_benchmark) 
    {
        setpriority(PRIO_PROCESS, 0, 19);
        drop_policy();
    }
    scratchbuf = scrypt_buffer_alloc(opt_scrypt_n);
    if (!scratchbuf) 
    {
        applog(LOG_ERR, "scrypt buffer allocation failed");
        pthread_mutex_lock(&applog_lock);
        exit(1);
    }

restart:
    while(sfards_port_open(sfards_dev,thr_id) < 0 && !sfards_dev->exit)
    {
        sleep(5);
    }

    stat_to = time(NULL) + STATISTC_INTERVAL;

    while (!sfards_dev->exit) 
    {
        if (workfifo_fetch(&workfifo,&work) != 0) 
        {
            sleep(1);
            stat_to++;
            continue;
        }
        
        work_restart[thr_id].restart = 0;
        
        if (sfards_dispatch_scrypt_task(port,work.target,work.midstate,work.data) < 0)
        {
            break;
        }

        while(!work_restart[thr_id].restart) 
        {
            clock_gettime(CLOCK_REALTIME, &ts);
            ts.tv_sec++;
            count = sfards_wait_nonce(port,nonce_rpt,8,&ts);
            for (i = 0;i < count;i++)
            {
                uint32_t nonce = nonce_rpt[i].data;
                int chipid = nonce_rpt[i].chipid;
                if (scrypt_verify_nonce(work.target,work.midstate,work.data,
                                        scratchbuf,opt_scrypt_n,nonce))
                {
                    stratum_submit_work(mythr,&work,0,nonce);
                    applog(LOG_INFO,"Got nonce %d %8.8x [%s]",chipid,nonce,port->path);
                    sfards_port_status_update(port,chipid,0,nonce,0);
                    sfards_send_devs_msg(port,sfards_dev->port_map,accepted_count,rejected_count,
                                         time(NULL) - start_time);
                }
                else
                {
                    applog(LOG_INFO,"Invalid nonce %d %8.8x [%s]",chipid,nonce,port->path);
                    sfards_port_status_update(port,chipid,0,nonce,1);
                }                
            }
        }
        if (stat_to < time(NULL))
        {
            stratum_port_status(port);

            if (sfards_scrypt_control(port) < 0)
            {
                break;
            }
            stat_to = time(NULL) + STATISTC_INTERVAL;
        }
    }

    sfards_port_close(port);

    if (!sfards_dev->exit)
    {
        applog(LOG_ERR,"Port restarting... (%s)",port->path);
        goto restart;
    }

    tq_freeze(mythr->q);

    return NULL;
}

static void *stratum_sha256d_thread(void *userdata)
{
    struct thr_info *mythr = userdata;
    struct sfards_port *port;
    struct work work[4];
    struct sfards_rpt nonce_rpt[8];
    struct timespec ts;
    time_t stat_to;
    uint32_t ntime0 = 0;
    uint32_t offset = 0;
    int thr_id;
    int taskid = 0;
    int i,count;
    
    thr_id = mythr->id;
    port = &sfards_dev->port[thr_id];

    memset(work,0,sizeof(struct work) * 4);

    /* Set worker threads to nice 19 and then preferentially to SCHED_IDLE
     * and if that fails, then SCHED_BATCH. No need for this to be an
     * error if it fails */
    if (!opt_benchmark) 
    {
        setpriority(PRIO_PROCESS, 0, 19);
        drop_policy();
    }

restart:
    while(sfards_port_open(sfards_dev,thr_id) < 0 && !sfards_dev->exit)
    {
        sleep(5);
    }

    stat_to = time(NULL) + STATISTC_INTERVAL;

    while(!sfards_dev->exit)
    {
        if (offset == 0 || work_restart[thr_id].restart)
        {
            if (workfifo_fetch(&workfifo,&work[taskid]) != 0) 
            {
                sleep(1);
                stat_to++;
                continue;
            }
            ntime0 = swab32(work[taskid].data[17]);
            work_restart[thr_id].restart = 0;
            offset = 0;
        }
        else
        {
            work_copy(&work[taskid],&work[(taskid - 1) & 3]);
            work[taskid].data[17] = swab32(ntime0 + offset);
        }
         
        clock_gettime(CLOCK_REALTIME, &ts);
        ts.tv_nsec += port->rolltime * 1000L;
        if (ts.tv_nsec > 1000000000L) 
        {
            ts.tv_sec++;
            ts.tv_nsec -= 1000000000L;
        }

        if (sfards_dispatch_sha256d_task(port,taskid,work[taskid].target,work[taskid].midstate,
                                         &work[taskid].data[16],ntime0 + offset) < 0)
        {
            break;
        }
        
        do
        {
            count = sfards_wait_nonce(port,nonce_rpt,8,&ts);
            for (i = 0;i < count;i++) 
            {
                struct work *cur = &work[nonce_rpt[i].taskid];
                uint32_t nonce = nonce_rpt[i].data;
                int chipid = nonce_rpt[i].chipid;
                if (sha256d_verify_nonce(cur->target,cur->midstate,
                                         &cur->data[16],chipid - 1,nonce)) 
                {
                    stratum_submit_work(mythr,cur,chipid - 1,nonce);
                    applog(LOG_INFO,"Got nonce %d(%d) %d %8.8x [%s]",
                                    nonce_rpt[i].taskid,taskid,chipid,nonce,port->path);
                    sfards_port_status_update(port,chipid,nonce_rpt[i].taskid,nonce,0);
                    sfards_send_devs_msg(port,sfards_dev->port_map,accepted_count,rejected_count,
                                         time(NULL) - start_time);
                }
                else
                {
                    applog(LOG_INFO,"Invalid nonce %d(%d) %d %8.8x [%s]",
                                    nonce_rpt[i].taskid,taskid,chipid,nonce,port->path);
                    sfards_port_status_update(port,chipid,nonce_rpt[i].taskid,nonce,1);
                }
            }
        }while(count > 0);

        taskid = (taskid + 1) & 3; 
        offset += 8;
        if (offset >= MAX_NTIME_OFFSET)
        {
            offset = 0;
        }

        if (stat_to < time(NULL))
        {            
            stratum_port_status(port);

            if (sfards_sha256d_control(port) < 0)
            {
                break;
            }
            
            stat_to = time(NULL) + STATISTC_INTERVAL;
        }
    }

    sfards_port_close(port);

    if (!sfards_dev->exit)
    {
        applog(LOG_ERR,"Port restarting... (%s)",port->path);
        goto restart;
    }

    tq_freeze(mythr->q);

    return NULL;
}

static void restart_threads(void)
{
    int i;

    for (i = 0; i < opt_n_threads; i++)
    {
        work_restart[i].restart = 1;
    }
}

static bool stratum_handle_response(char *buf)
{
    json_t *val, *err_val, *res_val, *id_val;
    json_error_t err;
    bool ret = false;

    val = JSON_LOADS(buf, &err);
    if (!val) 
    {
        applog(LOG_INFO, "JSON decode failed(%d): %s", err.line, err.text);
        goto out;
    }

    res_val = json_object_get(val, "result");
    err_val = json_object_get(val, "error");
    id_val = json_object_get(val, "id");

    if (!id_val || json_is_null(id_val) || !res_val)
        goto out;

    share_result(json_is_true(res_val),
                 err_val ? json_string_value(json_array_get(err_val, 1)) : NULL);

    ret = true;
out:
    if (val)
        json_decref(val);

    return ret;
}

static void *stratum_thread(void *userdata)
{
    struct thr_info *mythr = userdata;
    char *s;

    stratum.url = tq_pop(mythr->q, NULL);
    if (!stratum.url)
    {
        goto out;
    }
    applog(LOG_INFO, "Starting Stratum on %s", stratum.url);

    while (1) 
    {
        int failures = 0;

        while (!stratum.curl) 
        {
            restart_threads();

            if (!stratum_connect(&stratum, stratum.url) ||
                !stratum_subscribe(&stratum) ||
                !stratum_authorize(&stratum, rpc_user, rpc_pass)) 
            {
                stratum_disconnect(&stratum);
                if (opt_retries >= 0 && ++failures > opt_retries) 
                {
                    applog(LOG_ERR, "...terminating workio thread");
                    tq_push(thr_info[work_thr_id].q, NULL);
                    goto out;
                }
                applog(LOG_ERR, "...retry after %d seconds", opt_fail_pause);
                sleep(opt_fail_pause);
            }
        }

        if (strlen(stratum.job.job_id) && strcmp(stratum.job.job_id,workfifo.job_id)) 
        {
            strcpy(workfifo.job_id,stratum.job.job_id);
            workfifo_notify(&workfifo,WORKFIFO_NOTIFY_FLUSH);
            if (stratum.job.clean) 
            {
                applog(LOG_INFO, "Stratum requested work restart %s",stratum.job.job_id);
                restart_threads();
            }
            else
            {
                applog(LOG_INFO, "Stratum get new job %s",stratum.job.job_id);
            }
        }            
        
        if (!stratum_socket_full(&stratum, 120)) 
        {
            applog(LOG_ERR, "Stratum connection timed out");
            s = NULL;
        } 
        else
        {
            s = stratum_recv_line(&stratum);
        }
        
        if (!s) 
        {
            stratum_disconnect(&stratum);
            applog(LOG_ERR, "Stratum connection interrupted");
            continue;
        }
        if (!stratum_handle_method(&stratum, s))
        {
            stratum_handle_response(s);
        }
        free(s);
    }

out:
    return NULL;
}

static void show_version_and_exit(void)
{
    printf(PACKAGE_STRING "\n built on " __DATE__ "\n features:"
#if defined(USE_ASM) && defined(__i386__)
        " i386"
#endif
#if defined(USE_ASM) && defined(__x86_64__)
        " x86_64"
#endif
#if defined(USE_ASM) && (defined(__i386__) || defined(__x86_64__))
        " SSE2"
#endif
#if defined(__x86_64__) && defined(USE_AVX)
        " AVX"
#endif
#if defined(__x86_64__) && defined(USE_AVX2)
        " AVX2"
#endif
#if defined(__x86_64__) && defined(USE_XOP)
        " XOP"
#endif
#if defined(USE_ASM) && defined(__arm__) && defined(__APCS_32__)
        " ARM"
#if defined(__ARM_ARCH_5E__) || defined(__ARM_ARCH_5TE__) || \
    defined(__ARM_ARCH_5TEJ__) || defined(__ARM_ARCH_6__) || \
    defined(__ARM_ARCH_6J__) || defined(__ARM_ARCH_6K__) || \
    defined(__ARM_ARCH_6M__) || defined(__ARM_ARCH_6T2__) || \
    defined(__ARM_ARCH_6Z__) || defined(__ARM_ARCH_6ZK__) || \
    defined(__ARM_ARCH_7__) || \
    defined(__ARM_ARCH_7A__) || defined(__ARM_ARCH_7R__) || \
    defined(__ARM_ARCH_7M__) || defined(__ARM_ARCH_7EM__)
        " ARMv5E"
#endif
#if defined(__ARM_NEON__)
        " NEON"
#endif
#endif
#if defined(USE_ASM) && (defined(__powerpc__) || defined(__ppc__) || defined(__PPC__))
        " PowerPC"
#if defined(__ALTIVEC__)
        " AltiVec"
#endif
#endif
        "\n");

    printf("%s\n", curl_version());
#ifdef JANSSON_VERSION
    printf("libjansson %s\n", JANSSON_VERSION);
#endif
    exit(0);
}

static void show_usage_and_exit(int status)
{
    if (status)
        fprintf(stderr, "Try `" PROGRAM_NAME " --help' for more information.\n");
    else
        printf(usage);
    exit(status);
}

static void strhide(char *s)
{
    if (*s) *s++ = 'x';
    while (*s) *s++ = '\0';
}

static void parse_config(json_t *config, char *pname, char *ref);

static void parse_arg(int key, char *arg, char *pname)
{
    char *p;
    int v, i;

    switch(key) 
    {
    case 'd':
        if (strlen(sfards_paths) == 0)
        {
            strcpy(sfards_paths,arg);
        }
        else if (strlen(sfards_paths) + strlen(arg) < 253)
        {
            strcat(sfards_paths,",");
            strcat(sfards_paths,arg);
        }
        break;
    case 'f':
        sfards_freq = atoi(arg);
        break;
    case 'a':
        for (i = 0; i < ARRAY_SIZE(algo_names); i++) 
        {
            v = strlen(algo_names[i]);
            if (!strncmp(arg, algo_names[i], v)) 
            {
                if (arg[v] == '\0') 
                {
                    opt_algo = i;
                    break;
                }
                if (arg[v] == ':' && i == ALGO_SCRYPT) 
                {
                    char *ep;
                    v = strtol(arg+v+1, &ep, 10);
                    if (*ep || v & (v-1) || v < 2)
                        continue;
                    opt_algo = i;
                    opt_scrypt_n = v;
                    break;
                }
            }
        }
        if (i == ARRAY_SIZE(algo_names)) 
        {
            fprintf(stderr, "%s: unknown algorithm -- '%s'\n",
                pname, arg);
            show_usage_and_exit(1);
        }
        break;
    case 'B':
        opt_background = true;
        break;
    case 'c': 
        {
            json_error_t err;
            json_t *config = JSON_LOAD_FILE(arg, &err);
            if (!json_is_object(config)) 
            {
                if (err.line < 0)
                    fprintf(stderr, "%s: %s\n", pname, err.text);
                else
                    fprintf(stderr, "%s: %s:%d: %s\n",
                                   pname, arg, err.line, err.text);
                exit(1);
            }
            parse_config(config, pname, arg);
            json_decref(config);
            break;
        }
    case 'q':
        opt_quiet = true;
        break;
    case 'D':
        opt_debug = true;
        break;
    case 'p':
        free(rpc_pass);
        rpc_pass = strdup(arg);
        strhide(arg);
        break;
    case 'P':
        opt_protocol = true;
        break;
    case 'r':
        v = atoi(arg);
        if (v < -1 || v > 9999)    /* sanity check */
            show_usage_and_exit(1);
        opt_retries = v;
        break;
    case 'R':
        v = atoi(arg);
        if (v < 1 || v > 9999)    /* sanity check */
            show_usage_and_exit(1);
        opt_fail_pause = v;
        break;
    case 's':
        v = atoi(arg);
        if (v < 1 || v > 9999)    /* sanity check */
            show_usage_and_exit(1);
        opt_scantime = v;
        break;
    case 'T':
        v = atoi(arg);
        if (v < 1 || v > 99999)    /* sanity check */
            show_usage_and_exit(1);
        opt_timeout = v;
        break;
    case 't':
        v = atoi(arg);
        if (v < 1 || v > 9999)    /* sanity check */
            show_usage_and_exit(1);
        opt_n_threads = v;
        break;
    case 'u':
        free(rpc_user);
        rpc_user = strdup(arg);
        break;
    case 'o': {            /* --url */
        char *ap, *hp;
        ap = strstr(arg, "://");
        ap = ap ? ap + 3 : arg;
        hp = strrchr(arg, '@');
        if (hp) 
        {
            *hp = '\0';
            p = strchr(ap, ':');
            if (p) 
            {
                free(rpc_userpass);
                rpc_userpass = strdup(ap);
                free(rpc_user);
                rpc_user = calloc(p - ap + 1, 1);
                strncpy(rpc_user, ap, p - ap);
                free(rpc_pass);
                rpc_pass = strdup(++p);
                if (*p) *p++ = 'x';
                v = strlen(hp + 1) + 1;
                memmove(p + 1, hp + 1, v);
                memset(p + v, 0, hp - p);
                hp = p;
            } 
            else 
            {
                free(rpc_user);
                rpc_user = strdup(ap);
            }
            *hp++ = '@';
        } 
        else
        {
            hp = ap;
        }

        if (ap != arg) 
        {
            if (strncasecmp(arg, "http://", 7) &&
                strncasecmp(arg, "https://", 8) &&
                strncasecmp(arg, "stratum+tcp://", 14) &&
                strncasecmp(arg, "stratum+tcps://", 15)) 
            {
                fprintf(stderr, "%s: unknown protocol -- '%s'\n",
                    pname, arg);
                show_usage_and_exit(1);
            }
            free(rpc_url);
            rpc_url = strdup(arg);
            strcpy(rpc_url + (ap - arg), hp);
        } 
        else 
        {
            if (*hp == '\0' || *hp == '/') 
            {
                fprintf(stderr, "%s: invalid URL -- '%s'\n",
                    pname, arg);
                show_usage_and_exit(1);
            }
            free(rpc_url);
            rpc_url = malloc(strlen(hp) + 8);
            sprintf(rpc_url, "http://%s", hp);
        }
        have_stratum = !opt_benchmark && !strncasecmp(rpc_url, "stratum", 7);
        break;
    }
    case 'O':            /* --userpass */
        p = strchr(arg, ':');
        if (!p) 
        {
            fprintf(stderr, "%s: invalid username:password pair -- '%s'\n",
                pname, arg);
            show_usage_and_exit(1);
        }
        free(rpc_userpass);
        rpc_userpass = strdup(arg);
        free(rpc_user);
        rpc_user = calloc(p - arg + 1, 1);
        strncpy(rpc_user, arg, p - arg);
        free(rpc_pass);
        rpc_pass = strdup(++p);
        strhide(p);
        break;
    case 'x':            /* --proxy */
        if (!strncasecmp(arg, "socks4://", 9))
            opt_proxy_type = CURLPROXY_SOCKS4;
        else if (!strncasecmp(arg, "socks5://", 9))
            opt_proxy_type = CURLPROXY_SOCKS5;
#if LIBCURL_VERSION_NUM >= 0x071200
        else if (!strncasecmp(arg, "socks4a://", 10))
            opt_proxy_type = CURLPROXY_SOCKS4A;
        else if (!strncasecmp(arg, "socks5h://", 10))
            opt_proxy_type = CURLPROXY_SOCKS5_HOSTNAME;
#endif
        else
            opt_proxy_type = CURLPROXY_HTTP;
        free(opt_proxy);
        opt_proxy = strdup(arg);
        break;
    case 1001:
        free(opt_cert);
        opt_cert = strdup(arg);
        break;
    case 1005:
        opt_benchmark = true;
        want_longpoll = false;
        want_stratum = false;
        have_stratum = false;
        break;
    case 1003:
        want_longpoll = false;
        break;
    case 1007:
        want_stratum = false;
        break;
    case 1009:
        opt_redirect = false;
        break;
    case 1010:
        allow_getwork = false;
        break;
    case 1011:
        have_gbt = false;
        break;
    case 1013:            /* --coinbase-addr */
        pk_script_size = address_to_script(pk_script, sizeof(pk_script), arg);
        if (!pk_script_size) 
        {
            fprintf(stderr, "%s: invalid address -- '%s'\n",
                pname, arg);
            show_usage_and_exit(1);
        }
        break;
    case 1015:            /* --coinbase-sig */
        if (strlen(arg) + 1 > sizeof(coinbase_sig)) 
        {
            fprintf(stderr, "%s: coinbase signature too long\n", pname);
            show_usage_and_exit(1);
        }
        strcpy(coinbase_sig, arg);
        break;
    case 'S':
        use_syslog = true;
        break;
    case 'V':
        show_version_and_exit();
    case 'h':
        show_usage_and_exit(0);
    default:
        show_usage_and_exit(1);
    }
}

static void parse_config(json_t *config, char *pname, char *ref)
{
    int i;
    char *s;
    json_t *val;

    for (i = 0; i < ARRAY_SIZE(options); i++) 
    {
        if (!options[i].name)
            break;

        val = json_object_get(config, options[i].name);
        if (!val)
            continue;

        if (options[i].has_arg && json_is_string(val)) 
        {
            if (!strcmp(options[i].name, "config")) 
            {
                fprintf(stderr, "%s: %s: option '%s' not allowed here\n",
                    pname, ref, options[i].name);
                exit(1);
            }
            s = strdup(json_string_value(val));
            if (!s)
                break;
            parse_arg(options[i].val, s, pname);
            free(s);
        } 
        else if (!options[i].has_arg && json_is_true(val)) 
        {
            parse_arg(options[i].val, "", pname);
        } 
        else 
        {
            fprintf(stderr, "%s: invalid argument for option '%s'\n",
                pname, options[i].name);
            exit(1);
        }
    }
}

static void parse_cmdline(int argc, char *argv[])
{
    int key;

    while (1) 
    {
#if HAVE_GETOPT_LONG
        key = getopt_long(argc, argv, short_options, options, NULL);
#else
        key = getopt(argc, argv, short_options);
#endif
        if (key < 0)
            break;

        parse_arg(key, optarg, argv[0]);
    }
    if (optind < argc) 
    {
        fprintf(stderr, "%s: unsupported non-option argument -- '%s'\n",
            argv[0], argv[optind]);
        show_usage_and_exit(1);
    }
}

#ifndef WIN32
static void signal_handler(int sig)
{
    switch (sig) 
    {
    case SIGHUP:
        applog(LOG_INFO, "SIGHUP received");
        break;
    case SIGINT:
        applog(LOG_INFO, "SIGINT received, exiting");
        exit(0);
        break;
    case SIGTERM:
        applog(LOG_INFO, "SIGTERM received, exiting");
        exit(0);
        break;
    }
}
#endif

int main(int argc, char *argv[])
{
    struct thr_info *thr;
    long flags;
    int i;

    rpc_user = strdup("");
    rpc_pass = strdup("");

    /* parse command line */
    parse_cmdline(argc, argv);

    if (!opt_benchmark && !rpc_url) 
    {
        fprintf(stderr, "%s: no URL supplied\n", argv[0]);
        show_usage_and_exit(1);
    }

    if (!rpc_userpass) 
    {
        rpc_userpass = malloc(strlen(rpc_user) + strlen(rpc_pass) + 2);
        if (!rpc_userpass)
            return 1;
        sprintf(rpc_userpass, "%s:%s", rpc_user, rpc_pass);
    }

    pthread_mutex_init(&applog_lock, NULL);
    pthread_mutex_init(&stats_lock, NULL);
    pthread_mutex_init(&stratum.sock_lock, NULL);
    pthread_mutex_init(&stratum.work_lock, NULL);

    workfifo_init(&workfifo);

    flags = opt_benchmark || (strncasecmp(rpc_url, "https://", 8) &&
                              strncasecmp(rpc_url, "stratum+tcps://", 15))
          ? (CURL_GLOBAL_ALL & ~CURL_GLOBAL_SSL)
          : CURL_GLOBAL_ALL;
    if (curl_global_init(flags)) 
    {
        applog(LOG_ERR, "CURL initialization failed");
        return 1;
    }

#ifndef WIN32
    if (opt_background) 
    {
        i = fork();
        if (i < 0) exit(1);
        if (i > 0) exit(0);
        i = setsid();
        if (i < 0)
            applog(LOG_ERR, "setsid() failed (errno = %d)", errno);
        i = chdir("/");
        if (i < 0)
            applog(LOG_ERR, "chdir() failed (errno = %d)", errno);
        signal(SIGHUP, signal_handler);
        signal(SIGINT, signal_handler);
        signal(SIGTERM, signal_handler);
    }
#endif

#if defined(WIN32)
    SYSTEM_INFO sysinfo;
    GetSystemInfo(&sysinfo);
    num_processors = sysinfo.dwNumberOfProcessors;
#elif defined(_SC_NPROCESSORS_CONF)
    num_processors = sysconf(_SC_NPROCESSORS_CONF);
#elif defined(CTL_HW) && defined(HW_NCPU)
    int req[] = { CTL_HW, HW_NCPU };
    size_t len = sizeof(num_processors);
    sysctl(req, 2, &num_processors, &len, NULL, 0);
#else
    num_processors = 1;
#endif
    if (num_processors < 1)
        num_processors = 1;
    if (!opt_n_threads)
        opt_n_threads = num_processors;

    if (sfards_paths[0] != 0)
    {
        sfards_dev = sfards_new(opt_algo == ALGO_SCRYPT ?
                                PORT_TYPE_SCRYPT : PORT_TYPE_SHA256);
        if (sfards_dev != NULL)
        {
            char *p = sfards_paths;
            char *q,*s;
            int id,freq;
            int n = 0;
            do
            {
                id = 0;
                freq = sfards_freq;
                q = strchr(p, ',');
                if (q != NULL)
                {
                    *q = '\0';
                }

                s = strchr(p,':');
                if (s != NULL)
                {
                    *s = '\0';
                    id = atoi(p);
                    p = s + 1; 
                }
                s = strchr(p,'@');
                if (s != NULL)
                {
                    *s = '\0';
                    freq =  atoi(s + 1);
                }

                if (sfards_port_addnew(sfards_dev,n,p,id,freq) < 0)
                {
                    break;
                }
                n++;
                p = q + 1;
            }while(q != NULL);
            opt_n_threads = n;
        }
    }

#ifdef HAVE_SYSLOG_H
    if (use_syslog)
        openlog("cpuminer", LOG_PID, LOG_USER);
#endif

    work_restart = calloc(opt_n_threads, sizeof(*work_restart));
    if (!work_restart)
        return 1;

    thr_info = calloc(opt_n_threads + 4, sizeof(*thr));
    if (!thr_info)
        return 1;
    
    thr_hashrates = (double *) calloc(opt_n_threads, sizeof(double));
    if (!thr_hashrates)
        return 1;

    /* init workio thread info */
    work_thr_id = opt_n_threads;
    thr = &thr_info[work_thr_id];
    thr->id = work_thr_id;
    thr->q = tq_new();
    if (!thr->q)
        return 1;

    /* start work I/O thread */
    if (pthread_create(&thr->pth, NULL, workio_thread, thr)) 
    {
        applog(LOG_ERR, "workio thread create failed");
        return 1;
    }

    /* init workfifo thread info */
    workfifo_thr_id = opt_n_threads + 1;
    thr = &thr_info[workfifo_thr_id];
    thr->id = workfifo_thr_id;
    thr->q = tq_new();
    if (!thr->q)
        return 1;

    /* start work fifo thread */
    if (pthread_create(&thr->pth, NULL, workfifo_thread, thr)) 
    {
        applog(LOG_ERR, "workfifo thread create failed");
        return 1;
    }

    if (want_stratum) 
    {
        /* init stratum thread info */
        stratum_thr_id = opt_n_threads + 3;
        thr = &thr_info[stratum_thr_id];
        thr->id = stratum_thr_id;
        thr->q = tq_new();
        if (!thr->q)
            return 1;

        /* start stratum thread */
        if (unlikely(pthread_create(&thr->pth, NULL, stratum_thread, thr))) 
        {
            applog(LOG_ERR, "stratum thread create failed");
            return 1;
        }

        if (have_stratum)
            tq_push(thr_info[stratum_thr_id].q, strdup(rpc_url));
    }

    /* start mining threads */
    for (i = 0; i < opt_n_threads; i++) 
    {
        thr = &thr_info[i];

        thr->id = i;
        thr->q = tq_new();
        if (!thr->q)
            return 1;
        if (sfards_dev) 
        {
            if (opt_algo == ALGO_SHA256D) 
            {
                if (unlikely(pthread_create(&thr->pth, NULL, stratum_sha256d_thread, thr))) 
                {
                    applog(LOG_ERR, "thread %d create failed", i);
                    return 1;
                }
            }
            else if (unlikely(pthread_create(&thr->pth, NULL, stratum_scrypt_thread, thr))) 
            {
                applog(LOG_ERR, "thread %d create failed", i);
                return 1;
            }
        }
    }

    applog(LOG_INFO, "%d miner threads started, "
                     "using '%s' algorithm.",
                     opt_n_threads,
                     algo_names[opt_algo]);

    start_time = time(NULL);

    /* main loop - simply wait for workio thread to exit */
    pthread_join(thr_info[work_thr_id].pth, NULL);
    
    workfifo_notify(&workfifo,WORKFIFO_NOTIFY_EXIT);
    pthread_join(thr_info[workfifo_thr_id].pth, NULL);    
    workfifo_destroy(&workfifo);
//_exit:
    if (sfards_dev)
    {
        sfards_destroy(sfards_dev);
    }

    applog(LOG_INFO, "workio thread dead, exiting.");

    return 0;
}


